#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from __future__ import print_function

#%% 

import pandas as pd
import numpy as np
import operator


from sklearn.model_selection import train_test_split
from datetime import datetime



##
## loadData
## Version: 0.0.1
## 
## Descripcion: Se pasa un fichero en formato txt y se devuelve el resultado
## en formato dataframe ordenado por Id del comprador y la fecha.
##
## Entrada:
## - filename: nombre del fichero en formato txt.
##
## Salida:
## - table: descripcion.
##
## Modificaciones:
##
def loadData (filename):
    
    df = pd.read_csv(filename,delimiter='|')
    df = df.sort_values(by=['ID_Customer', 'Cod_Fecha'], ascending=[1, 1]).reset_index(drop=True)
    return df

##
## dfToFile
## Version: 0.0.1
## 
## Descripcion: Vuelca un dataframe a un fichero.
##
## Entrada:
## - df: dataframe que vamos a pasar a un fichero
## - file_name : nombre del fichero de salida
##
## Salida:
##
## Modificaciones: 
##
def dfToFile(df,file_name):
    
    df.to_csv(file_name, sep='|', encoding='utf-8',index=False)
    
    return
##
## getDfMap
## Version: 0.0.1
## 
## Descripcion: esta funcion recibe el diccionario de train
## y el dataframe final y mapea los codigo de producto en el 
## formato de salida
##
## Entrada:
## - df: dataframe con ID_Customer y codigo de producto predicho
## - dictTrain : diccionario que hemos obtenido de entrenamiento
##
## Salida:
## - df: con el mapeado realizado
##
## Modificaciones: 
##    
def getDfMap(df,dictTrain=[]):      
    if dictTrain == []:
        dictTrain = mapProduct.mapPrioris
        
    dictFinal={}
    for idGenerado in dictTrain.keys():
        dictFinal[int(dictTrain[idGenerado])]= str(idGenerado).zfill(4)
    
    df = df.replace({"Cod_Prod":dictFinal})
    
    return df



##
## tratamientoFecha
## Version: 0.0.1
## 
## Descripcion: Se pasa un dataframe y se crea una nueva columna con el numero 
## de dias que han pasado desde 1950-01-1 hasta que ese usuario ha hecho la 
## compra del producto
## 
##
## Entrada:
## - df: dataframe al cual vamos a crear la columna numero de dias.
##
## Salida:
## - dfFecha: devolvemos el df con la nueva columna num_dias.
##
## Modificaciones:
##    
def tratamientoFecha(df):
    
    dfFecha=df.copy()
    
    #Pasamos la columna Cod_Fecha a formato año y mes
    dfFecha["Cod_Fecha"] =  pd.to_datetime(dfFecha["Cod_Fecha"],format='%Y-%m')
    primerMes = pd.datetime(1950, 1, 1, 0, 0, 0)
    
    #Realizamos la resta del numero de dias desde 1950 y creamos la columna
    dfFecha["Num_Dias"] =  dfFecha["Cod_Fecha"]-primerMes 
    dfFecha["Num_Dias"] = (dfFecha["Num_Dias"] / np.timedelta64(1, 'D')).astype(int)
    
    return dfFecha

##
## mapProduct
## Version: 0.0.7
## 
## Descripcion: se pasa un dataframe de train y se obtiene el mapeado de productos
## ordenado por priori en caso de pasar la variable prioris a True.
##
## Entrada:
## - df: dataframe del cual obtendremos los ids de producto
##
## Salida:
## - dfMap: dataframe con el mapeado aplicado.
##
## Modificaciones:
##    0.0.2 ordenamos por fecha
##    0.0.3 añadimos el diccionario a producto
##    0.0.4 pasamos a tiempo unix mejor  
##    0.0.5 se añade socio_demo_05 como nominal y crea diccionario de los nominales
##    0.0.6 se quita la ordenación por fecha se añaden argumentos voluntarios y 
##          se incluye el mapeado por priori
##    0.0.7 Quitamos el tiempo unix
##

    
def mapProduct (df, prioris=True):
    dfMap=df.copy()
    mapProduct.mapProd = {}
    mapProduct.mapPrioris = {}
    
    
    if(prioris):
        dfPrioris= (dfMap.groupby("Cod_Prod").count().ID_Customer)/float(len(dfMap))
        
        dfPrioris = dfPrioris.sort_values(ascending = False)
        dfPrioris = dfPrioris.to_frame()
        
        dfPrioris.columns = ['Priori']
        dfPrioris["Labels"] = list(range(len(dfPrioris)))
        
        
        mapProduct.mapPrioris = pd.Series(dfPrioris.Labels.values,index=dfPrioris.index).to_dict()
        
        dfMap = dfMap.replace({"Cod_Prod": mapProduct.mapPrioris})
        
           
    else:
        #pasamos las columnas a datos categoricos
        dfMap['Cod_Prod'] = dfMap['Cod_Prod'].astype('category')
        #Extraemos las columnas con datos categoricos
        cat_columns = dfMap.select_dtypes(['category']).columns
        #Diccionario de IdProd : nombreProd
        map_prod = dict(enumerate(dfMap.Cod_Prod.cat.categories))
        
        #Los transformamos a numeros
        dfMap[cat_columns] = dfMap[cat_columns].apply(lambda x: x.cat.codes)
        
        for idGenerado in map_prod.keys():
            mapProduct.mapProd[int(map_prod[idGenerado])]=int(idGenerado)
    
    return dfMap

##
## mapProdByDict
## Version: 0.0.1
## 
## Descripcion: Mapea los codigo de producto de un dataframe con el diccionario
## pasado por argumento.
##
## Entrada:
## - df: dataframe que contiene la columna cod_prod para mapearla.
## - mapTest: diccionario con el mapeado del codigo de producto
##
## Salida:
## - df: dataframe con el mapeado aplicado.
##
## Modificaciones:
##

def mapProdByDict(df,mapTest = []):      
    if mapTest == []:
        mapTest = mapProduct.mapPrioris
        
    df = df.replace({"Cod_Prod":mapTest})
    
    return df

##
## addProdAnt
## Version: 0.0.1
## 
## Descripcion: funcion que extrae como nuevas columnas los productos
## anteriores comprados por el usuario, los dias que han pasado desde que se
## han comprado los X productos anteriores y la fecha del anterior producto.
##
## Entrada:
## - df: dataframe del cual vamos a extraer la informacion mencionada en la 
##  descripcion de la funcion.
## - num_ant: numero de productos anteriores a tener en cuenta
## - diasAnt: si pasamos esta variable tendremos en cuenta tambien los dias que
##   han pasado desde que se han comprado los X productos anteriores
## - fechaAnt1: si agregamos esta variable a true añadimos la fecha anterior del
##  producto anterior contratado por el cliente
##
## Salida:
## - df2: dataframe normalizado.
##
## Modificaciones:
##    
def addProdAnt(df,num_ant = 1, diasAnt = False, fechaAnt1 = True):
    
    df2 = df.copy().reset_index(drop=True)
    
    
    for ant in range(num_ant):
        
        if ant == 0:
            dfaux = list(df2.Cod_Prod)
            dfFechaAnt1 = list(df2.Cod_Fecha)
            if diasAnt:
                dffec = list(df2.Num_Dias)
        else:
            dfaux = list(df2["Cod_Prod_Ant"+str(ant)])
            if diasAnt:
                dffec = list(df2["Num_Dias_Ant"+str(ant)])

        anteriores = list(df2.groupby("ID_Customer",as_index=False).count().Cod_Prod)
        anteriores.pop(len(anteriores)-1)
        anteriores = np.cumsum(anteriores)
        
        dfaux.insert(0,-1)
        dfaux.pop(len(dfaux)-1)
        dfaux = np.array(dfaux)
        dfaux[anteriores]=-1
        
        df2["Cod_Prod_Ant"+str(ant+1)] = pd.DataFrame(dfaux)
        
        #Si tenemos decidimos añadir la columna de la fecha anterior
        if fechaAnt1 and ant==0:
            dfFechaAnt1.insert(0,datetime(1950, 1, 1, 0, 0, 0))
            dfFechaAnt1.pop(len(dfFechaAnt1)-1)
            dfFechaAnt1 = np.array(dfFechaAnt1)
            dfFechaAnt1[anteriores]=datetime(1950, 1, 1, 0, 0, 0)
            df2["Cod_Fecha_Ant"] = pd.DataFrame(dfFechaAnt1)
            
        #Si tenemos decidimos añadir la columna de los dias anteriores
        if diasAnt:
            dffec.insert(0,0)
            dffec.pop(len(dffec)-1)
            dffec = np.array(dffec)
            dffec[anteriores]=0
            
            df2["Num_Dias_Ant"+str(ant+1)] = pd.DataFrame(dffec)

    return df2

##
## createTest
## Version: 0.0.1
## Libreria: BLcleansing
## 
## Descripcion: esta funcion crea el dataframe de test con un registro por usuario
## 
##
## Entrada:
## - df: dataframe del cual extreremos el test.
##
## Salida:
## - dfTest: dataframe con un registro por usuario.
##
## Modificaciones:
##  
def createTest (df):
    dfTest2 = df.copy()
    
    #realizamos un group by por usuario y nos quedamos con el ultimo registro de 
    #cada uno
    dfTest2 = dfTest2.groupby("ID_Customer").last().reset_index()
    dfTest = dfTest2.copy()
    
    #Extraemos las columnas que debemos modificar para obtener informacion
    # coherente respecto a los productos anteriores
    listaColumnas = list(dfTest.columns)
    
    numProdAnt  = len(list(filter(lambda x: 'Cod_Prod_Ant' in x, listaColumnas)))
   
    dfTest["Cod_Fecha_Ant"]=dfTest2["Cod_Fecha"]
    
    #Modificamos las columnas anteriormente extraidas
    for i in range(numProdAnt):
        nameColProd= 'Cod_Prod_Ant'+str(i+1)
        nameColDias= 'Num_Dias_Ant'+str(i+1)
        
        nameColProd2= 'Cod_Prod_Ant'+str(i)
        nameColDias2= 'Num_Dias_Ant'+str(i)
        if i == 0:
            
            dfTest[nameColProd]=dfTest2["Cod_Prod"]
            dfTest[nameColDias]=dfTest2["Num_Dias"]
         
        else:
            dfTest[nameColProd]=dfTest2[nameColProd2]
            dfTest[nameColDias]=dfTest2[nameColDias2]
        

    return dfTest


##
## subset
## Version: 0.0.1
## 
## Descripcion: extrae un submuestreo del dataset original.
## 
## Entrada:
## - dfData: dataframe con los datos
## - dfLabels: dataframe con la clase
## - size: porcentaje que queremos extraer de los datos
## - seed: semilla para hacer el submuestreo
##
## Salida:
## - X_train: dataframe la reduccion aplicada.
##
## Modificaciones:
##  
def subset(dfData, dfLabels, size=.01, seed=45):
    X_train, X_test, y_train, y_test = train_test_split( dfData, dfLabels, train_size=size, random_state=seed)
    return X_train


##
## classPrune2
## Version: 0.0.1
## 
## Descripcion: Elimina clases dado un numero minimo de clases
##
## Entrada:
## - dfData: dataframe con los datos
## - dfLabels: dataframe con la clase
## - nc: numero de clases
## - resto: si esta a True añade el resto de clase a una variable y pone la etiqueta
##   de la clase mayoritaria a ese conjunto
##
## Salida:
## - df: dataframe con la poda realizada.
##
## Modificaciones:
##  
def classPrune2(dfData, dfLabels, nc = 60, resto = False):
    df = dfData.copy()
    count = dfLabels.value_counts()
    count = pd.DataFrame([count.index, count, count/count.sum()]).transpose()
    quedo = count[[0,1,2]][0:nc]
    poda = count[[0,1,2]][nc:len(count)]
    
    print ("Numero de clases:", nc,"\nPorcentaje datos:",quedo[2].sum()*100,"%\nMínimo de muestras:", min(quedo[1]))

    if not resto:
        ind = []
        for d in np.array(poda[0]):
            ind = np.concatenate([ind,df.index[(dfLabels==d)]],axis=0)
        df.drop(ind,axis=0,inplace=True)
    else:
        ind = []
        for d in np.array(poda[0]):
            ind = np.concatenate([ind,df.index[(dfLabels==d)]],axis=0)
        majClass = (np.array(poda)[0,0])
        for i in ind:
            df.loc[i,"Cod_Prod"] = majClass
    return df

##
## expandirVariable
## Version: 0.0.1
## 
## Descripcion: Expande la variable pasada por argumento
##
## Entrada:
## - df: dataframe en formato de recomendacion
## - nameCol: columna a expandir
##
## Salida:
##
## Modificaciones: 
##
def expandirVariable(df, nameCol):
    dfNuevo = df.copy()
    numVal = len(dfNuevo[nameCol].unique())
    print("Expandiendo "+ nameCol + " a dimension " + str(numVal))
    
    for i in range(numVal):
        nameColNuevo = nameCol + "_" + str(i+1)
        
        dfNuevo[nameColNuevo] = np.where(dfNuevo[nameCol]==i, 1, 0)
        
    dfNuevo = dfNuevo.drop(nameCol, 1)
    return dfNuevo
        

##
## mapAparicionProd
## Version: 0.0.1
## 
## Descripcion: crea un diccionario con el id del proucto y su orden de aparicion
##
## Entrada:
## - df: dataframe para extraer la informacion de la fecha de aparicion
## - productos: lista de todos los productos
##
## Salida:
##  fechaIDtemp: diccionario con la fecha y el orden en la que aparece cada producto
##    
## Modificaciones: 
##
def mapAparicionProd(df,productos = []):

    if productos == []:
        productos = list(mapProduct.mapPrioris.values())
        
    mapAparicion = {}
    
    for idP in productos:
        mapAparicion[idP]=df.loc[df['Cod_Prod'] == idP].sort_values(by = ["Cod_Fecha"])["Cod_Fecha"].iloc[0]
    sortedMap = sorted(mapAparicion.items(), key=operator.itemgetter(1))

    marcaTemporal = 0
    mapaFinal = {}
    
    #creamos el mapa
    for idP,fecha in sortedMap:
         mapaFinal[pd.to_datetime(fecha)] = marcaTemporal 
         marcaTemporal += 1
         
    fechaIDtemp = sorted(mapaFinal.items(), key=operator.itemgetter(1))    
         
    mapAparicionProd.mapAparicion = fechaIDtemp
    return fechaIDtemp

##
## addAparicionProdRow
## Version: 0.0.1
## 
## Descripcion: introduce la marca temporal de aparicion de cada producto en una fila
##
## Entrada:
## - row: fila a en la que vamos a añadir el valor
##
## Salida:
## - tiempoAnt: valor que va a tomar la fila
##    
## Modificaciones: 
##
def addAparicionProdRow(row):
    tiempoAnt = -1
    
    for idF,tiempo in mapAparicionProd.mapAparicion:
        if idF <= row["Cod_Fecha_Ant"]:
            tiempoAnt = tiempo
        else:
            break;
                
        
    return tiempoAnt
##
## addTiempoProdAnt
## Version: 0.0.1
## 
## Descripcion: creamos una columna con la marca temporal de aparicion del producto,
## a partir de la fecha anterior
##
## Entrada:
## - dfData: dataframe en el que creamos la nueva columna
##
##
## Salida:
## - df: dataframe modificado
##    
## Modificaciones: 
##
def addTiempoProdAnt(dfData):
    df = dfData.copy()
    
    df["Aparicion_Prod"] = df.apply(addAparicionProdRow, axis=1)
    
    return df

##
## mapDiasInicio
## Version: 0.0.1
## 
## Descripcion: creamos un mapa con el Id del comprador y la fecha de compra del
## primer producto.
##
## Entrada:
## - df: dataframe del cual extraemos la información del Id del comprador y la fecha
##   de su primer producto comprado. 
##
##
## Salida:
## - mapUFecha: diccionario con el id del comprador y la fecha del primer producto
##  comprado.
##    
## Modificaciones: 
##
def mapDiasInicio(df):
    dfUsuarioFecha = df[["ID_Customer","Cod_Fecha"]].groupby(["ID_Customer"]).first().reset_index()
    mapUFecha = dict(zip(dfUsuarioFecha.ID_Customer, dfUsuarioFecha.Cod_Fecha))
    
    if hasattr(mapDiasInicio,'mapUsuarioFecha'):
        mapDiasInicio.mapUsuarioFecha.update(mapUFecha)
    else:
        mapDiasInicio.mapUsuarioFecha = mapUFecha
    return mapUFecha

##
## addDiasInicioRow
## Version: 0.0.1
## 
## Descripcion: funcion que devuelve los dias de desde que empezo a comprar el usuario
## 
## Entrada:
## - row: fila de la que extraeremos los datos
##
## Salida:
## - val: valor de los dias que han pasado desde que ha empezado a comprar el usuario
##    
## Modificaciones: 
##
def addDiasInicioRow(row):

    if row["Cod_Prod_Ant1"] == -1:
        val = 0
    else:
        val =  (row["Cod_Fecha_Ant"] - mapDiasInicio.mapUsuarioFecha[row["ID_Customer"]]).days
        
    return val


##
## addDiasInicioAnt
## Version: 0.0.1
## 
## Descripcion: creamos una columna con los dias desde que empieza a comprar el usuario
## a partir de la fecha anterior.
##
## Entrada:
## - dfData: dataframe en el que creamos la nueva columna
##
##
## Salida:
## - df: dataframe modificado
##    
## Modificaciones: 
##
def addDiasInicioAnt(dfData):
    df = dfData.copy()
    
    df["DiasDesde_Inicio"] = df.apply(addDiasInicioRow, axis=1)
    
    return df


##
## numProductosComprados
## Version: 0.0.1
## 
## Descripcion: Extrae el numero de productos comprados por usuario
##
## Entrada:
## - df: dataframe del cual extraemos los datos
## - test: si pasamos el test debemos restar 1 al numero de prod
##
##
## Salida:
## - nprods: dataframe con el numero de productos comprados
##    
## Modificaciones: 
##
def numProductosComprados(df, test = False):
    nprods = df.groupby("ID_Customer").Cod_Prod.count().reset_index()
    if test:
        return nprods["Cod_Prod"]
    else:
        return nprods["Cod_Prod"] - 1

##
## ultimoElementoSerie
## Version: 0.0.1
## 
## Descripcion: Extrae el ultimo elemento de la serie de cada usuario
##
## Entrada:
## - df: dataframe del cual extraemos los datos
## 
##
##
## Salida:
## - df2: dataframe para cada usuario su ultima serie
##    
## Modificaciones: 
##
def ultimoElementoSerie(df):
    df2 = df.copy()
    return df2.groupby("ID_Customer").last().reset_index()


##
## restaFechas
## Version: 0.0.1
## 
## Descripcion: creamos columnas con las diferencias de fechas de compra de los
## productos anteriores en dias.
##
## Entrada:
## - df: dataframe del cual extraemos los datos
## 
## Salida:
## - df2: dataframe con la resta de fechas aplicada
##    
## Modificaciones: 
##
def restaFechas(dfData):
    
    df = dfData.copy()
    listaColumnas = list(df.columns)
    numResta  = len(list(filter(lambda x: 'Num_Dias_Ant' in x, listaColumnas)))
    numResta = numResta-1
    for i in range(numResta):
        nameCol = "Diferencia_Fechas_"+str(i+1)
        columna1 = "Num_Dias_Ant"+str(i+1)
        columna2 = "Num_Dias_Ant"+str(i+2)
        df[nameCol] = df[columna1] - df[columna2]
    
    
    return df

##
## modificaColumna
## Version: 0.0.1
## 
## Descripcion: selecciona el acontecimiento de cajamar perteneciente a la fusion de
## cajas rurales en funcion de la fecha anterior
##
## Entrada:
## - row: fila con los datos que vamos a modificar
## 
## Salida:
## - val: valor que corresponde con un acontecimiento de la historia de cajamar
##    
## Modificaciones: 
##
def modificaColumna(row):
    fecha1 = datetime(2000, 1, 1, 0, 0, 0) #fusion cajas rurales de Andalucía y Madrid
    fecha2 = datetime(2007, 1, 1, 0, 0, 0) #fusion Castilla y León
    fecha3 = datetime(2011, 1, 1, 0, 0, 0) #fusion Comunidad Valenciana
    if row["Cod_Fecha_Ant"] == -1:
        val = -1
    elif row["Cod_Fecha_Ant"] < fecha1:
        val = 0
    elif row["Cod_Fecha_Ant"] > fecha1 and row["Cod_Fecha_Ant"] < fecha2:
        val = 1
    elif row["Cod_Fecha_Ant"] > fecha1 and row["Cod_Fecha_Ant"] > fecha2 and row["Cod_Fecha_Ant"] < fecha3 :
        val = 2
    else:
        val = 3
    return val
    
##
## acontecimiento
## Version: 0.0.1
## 
## Descripcion: añade el acontecimiento de cajamar perteneciente a la fusion de
## cajas rurales en funcion de la fecha anterior
##
## Entrada:
## - dfData: dataframe en el que creamos la nueva columna
## 
## Salida:
## - df: dataframe modificado
##    
## Modificaciones: 
##
def acontecimiento(dfData):
    df = dfData.copy()
    
    df["AcontecimientoAnt"] = df.apply(modificaColumna, axis=1)
    
    return df

##
## mapYearPIB
## Version: 0.0.1
## 
## Descripcion: crea un diccionario con el PIB per capita por año.
##
## Entrada:
## 
## Salida:
## - mapYearPIB.mapPIB: diccionario con el año asociado a su PIB per capita
##    
## Modificaciones: 
##
def mapYearPIB():
    dfPIB =  pd.read_csv("PIB.txt",delimiter='\t')

    mapYearPIB.mapPIB = pd.Series(dfPIB.PIB.values,index=dfPIB.Year).to_dict()
    
    return mapYearPIB.mapPIB

##
## addPIBRow
## Version: 0.0.1
## 
## Descripcion: selecciona el PIB per capita de la fecha de la anterior de compra
## de un cliente
##
## Entrada:
## - row: fila de la que extraemos la informacion  
## 
## Salida:
## - val: valor del valor per capita del año de la anterior compra
##    
## Modificaciones: 
##
def addPIBRow(row):
    
    try:
        if row["Cod_Fecha_Ant"].year == 1950: #si es 1950 no habia producto anterior
            val = -1
        elif row["Cod_Fecha_Ant"].year ==2017:
            val = 24000
        else:
            val = mapYearPIB.mapPIB[row["Cod_Fecha_Ant"].year]
            
    except KeyError: #si es menor de 1961 no tenemos datos del PIB asique lo ponemos a 0
        val = 147
    return val
##
## addPIBAnt
## Version: 0.0.1
## 
## Descripcion: selecciona el PIB per capita de la fecha de la anterior de compra
## de un cliente
##
## Entrada:
## - dfData: dataframe en el que creamos la nueva columna
## 
## Salida:
## - df: dataframe modificado
##    
## Modificaciones: 
##
def addPIBAnt(dfData):
    df = dfData.copy()
    
    df["PIB_Ant"] = df.apply(addPIBRow, axis=1)
    
    return df


